/*	dummyfill.c
 *
 *	Example program for v4l loopback device that works with v4l/v4l2.
 *	Based on the programs "invert" and "dummy" from the author of
 *	 "v4l loopback device" Jeroen Vreeken.
 *
 *	This program feeds a v4l loopback device with squares, lines, etc
 *	in format RGB24, trying to generate some kind of motion.
 *
 *	Copyright 2004 by Oliver Gonzalez (mallok@andaluciajunta.es)
 *
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <linux/videodev.h>

#define byte unsigned char
#define uint unsigned int

#define RGB24 3
#define SPEED 0.2f
#define GROW 0.01f
#define FORWARD SPEED
#define BACKWARD -SPEED
#define BIGGER GROW
#define SMALLER -GROW
#define SQUARE_BIG_MAX 100
#define SQUARE_BIG_MIN 10
#define BACKGROUND_COLOR 0
#define SCREEN_MAX_SQUARES 10
#define SCREEN_MIN_SQUARES 5

struct square {
    float x_pos;
    float y_pos;
    float big;
    uint colour;
    float big_inc;
    float x_direction;
    float y_direction;
};

// Number of squares to be painted
uint n_squares;

void initialize_random_seed() {
    uint seed;
    int handle;
    
    if ((handle = open("/dev/urandom", O_RDONLY | O_NONBLOCK, 0)) > -1) {
	// Using random device (better)
	read (handle, &seed, sizeof(uint));
	close (handle);
    } else {
	// Using the system local time
	struct timeval tv;
	struct timezone tz;
	gettimeofday (&tv, &tz);
	seed = tv.tv_usec;
    }
    srandom(seed);
}

uint get_random(uint max, uint min) {
    uint value;
    while ((value = random() % max) <= min);
    
    return value;
}

int start_pipe (int dev, uint width, uint height)
{
    struct video_capability vid_caps;
    struct video_window vid_win;
    struct video_picture vid_pic;

    // Getting video device capabilities
    if (ioctl (dev, VIDIOCGCAP, &vid_caps) == -1) {
	perror ("ioctl (VIDIOCGCAP)");
	return (1);
    }
    
    // Getting video device picture formats
    if (ioctl (dev, VIDIOCGPICT, &vid_pic) == -1) {
	perror ("ioctl VIDIOCGPICT");
	return (1);
    }
    
    // Setting video device format
    vid_pic.palette = VIDEO_PALETTE_RGB24;
    if (ioctl (dev, VIDIOCSPICT, &vid_pic) == -1) {
	perror ("ioctl VIDIOCSPICT");
	return (1);
    }
    
    // Gettin video device capture sizes
    if (ioctl (dev, VIDIOCGWIN, &vid_win) == -1) {
	perror ("ioctl VIDIOCGWIN");
	return (1);
    }
    
    vid_win.width = width;
    vid_win.height = height;
    
    // Setting video device capture size
    if (ioctl (dev, VIDIOCSWIN, &vid_win) == -1) {
	perror ("ioctl VIDIOCSWIN");
	return (1);
    }
	
    return 0;
}

void draw_line (byte *map, int x_init, int y_init, int x_final, int y_final, 
int e, int e_inc, int e_noinc, uint width, uint colour, int pred, int incdec) {
    
    // Decoding the colour
    byte *rgb = (byte *) &colour;

    int start, end, var;

    if (pred == 1) {
	start = x_init;
	end = x_final;
	var = y_init;
    }
    else {
	start = y_init;
	end = y_final; 
	var = x_init;
    }
    
    int i, offset;
    for (i = start; i <= end; i++) {
	if (pred == 0) {
	    offset = ((var * width) + i) * RGB24;
	    map[offset] = rgb[0]; // RED
	    map[offset + 1] = rgb[1]; // GREEN
	    map[offset + 2] = rgb[2]; // BLUE
	}
	else {
	    offset = ((i * width) + var) * RGB24;
	    map[offset] = rgb[0]; // RED
	    map[offset + 1] = rgb[1]; // GREEN
	    map[offset + 2] = rgb[2]; // BLUE
	}

	if (e < 0)
	    e += e_noinc;
	else {
	    var += incdec;
	    e += e_inc;
	}
    }
} 

// Bresenham algorithm
void bresenham_drawline(byte *map, uint x_init, uint y_init,
		uint x_final, uint y_final, uint width, uint colour) {

    // Decoding the colour
    byte *rgb = (byte *) &colour;

    int t, i, dx, dy, e, e_inc, e_noinc, offset;
    
    if (x_init > x_final) {
	t = x_init; x_init = x_final; x_final = t;
	t = y_init; y_init = y_final; y_final = t;
    }

    dx = x_final - x_init; dy = y_final - y_init;

    /* vertical line */
    if (dx == 0) { 
	if (y_init > y_final) {
	    t = y_init; y_init = y_final; y_final = t;
	}

	offset = ((i * width) + x_init) * RGB24;
	for (i = y_init; i <= y_final; i++) {
	    offset = ((i * width) + x_init) * RGB24;
	    map[offset] = rgb[0]; // RED
	    map[offset + 1] = rgb[1]; // GREEN
	    map[offset + 2] = rgb[2]; // BLUE
	}
	    
	return;
    }
    
    /* horizontal line */
    if (dy == 0) { 
	for (i = x_init; i < x_final; i++) {
	    offset = ((y_init * width) + i) * RGB24;
	    map[offset] = rgb[0]; // RED
	    map[offset + 1] = rgb[1]; // GREEN
	    map[offset + 2] = rgb[2]; // BLUE
	}
	    
        return;
    }

    /* 0 < m < 1 */
    if (dy < dx && dy > 0) {
	e_noinc = 2 * dy;
	e = 2 * dy - dx;
	e_inc = 2 * (dy - dx);
	draw_line (map, x_init, y_init, x_final, y_final, e, e_inc, e_noinc,
						 width, colour, 1, 1);
    }

    /* m = 1 */
    if (dy == dx && dy > 0) {
	e_noinc = 2 * dy;
	e = 2 * dy - dx;
	e_inc = 2 * (dy - dx);
	draw_line (map, x_init, y_init, x_final, y_final, e, e_inc, e_noinc,
						 width, colour, 1, 1);
    }

    /* 1 < m < infinity */
    if (dy > dx && dy > 0) {
	e_noinc = 2 * dx;
	e = 2 * dx - dy;
	e_inc = 2 * (dx - dy);
	draw_line (map, x_init, y_init, x_final, y_final, e, e_inc, e_noinc,
						 width, colour, 0, 1);
    }

    /* 0 > m > -1 */
    if (-dy < dx && dy < 0) {
	dy = -dy;
	e_noinc = 2 * dy;
	e = 2 * dy - dx;
	e_inc = 2 * (dy - dx);
	draw_line (map, x_init, y_init, x_final, y_final, e, e_inc, e_noinc,
						 width, colour, 1, -1);
    }

    /* m = -1 */
    if (dy == -dx && dy < 0) {
	dy = -dy;
	e_noinc = (2 * dy);
	e = 2 * dy - dx;
	e_inc = 2 * (dy - dx);
	draw_line (map, x_init, y_init, x_final, y_final, e, e_inc, e_noinc,
						 width, colour, 1, -1);
    }

    /* -1 > m > 0 */
    if (-dy > dx && dy < 0) {
	dx = -dx;
	e_noinc = - (2 * dx);
	e = 2 * dx - dy;
	e_inc = - 2 * (dx - dy);
	draw_line (map, x_final, y_final, x_init, y_init, e, e_inc, e_noinc,
						 width, colour, 0, -1);
    }
}

void draw_square(byte *map, uint x_init, uint y_init,
		uint x_final, uint y_final, uint width, uint colour) {

    // Draw a square by drawing lines
    int y;
    for (y = y_init; y < y_final; y++) 
	bresenham_drawline (map, x_init, y, x_final, y, width, colour);
}

void next_frame (byte *map, uint width, uint height, struct square *squares)
{
    uint x, y, big;
    
    // Setting the background to black
    memset (map, BACKGROUND_COLOR, width * height * RGB24);

    // Drawing squares
    int i;
    for (i = 0; i < n_squares; i++) {
	x = (uint) squares[i].x_pos;
	y = (uint) squares[i].y_pos;
	big = (uint) squares[i].big;

	// Checking boundaries
	if (big >= SQUARE_BIG_MAX) {
	    // Switches square direction
	    if ((get_random(2,0)) == 1)
		squares[i].x_direction = FORWARD;
	    else
		squares[i].x_direction = BACKWARD;

	    squares[i].big_inc = SMALLER;
	    big = SQUARE_BIG_MAX;
	}
	else if (big <= SQUARE_BIG_MIN) {
	    // Switches square direction
	    if ((get_random(2,0)) == 1)
		squares[i].y_direction = FORWARD;
	    else
		squares[i].y_direction = BACKWARD;

	    squares[i].big_inc = BIGGER;
	    big = SQUARE_BIG_MIN;
	}

	if (x >= (width - big)) {
	    squares[i].x_direction = BACKWARD;
	    x = (width - big);
	}
	else if (x <= 1) {
	    squares[i].x_direction = FORWARD;
	    x = 1;
	}
	    
	if (y >= (height - big)) {
	    squares[i].y_direction = BACKWARD;
	    y = (height - big);
	}
	else if (y <= 1) {
	    squares[i].y_direction = FORWARD;
	    y = 1;
	}
	    
	// Colour changing
	squares[i].colour++; 

	// Draw the square
	draw_square (map, x, y, x + big,
		     y + big, width, squares[i].colour);

	// Moving to any direction
	squares[i].x_pos += squares[i].x_direction;
	squares[i].y_pos += squares[i].y_direction;
	squares[i].big += squares[i].big_inc;
    }
}
		 
int put_image(int dev, byte *image, uint width, uint height)
{
    // Write the image to the v4l device
    if (write (dev, image, width * height * RGB24) != (width * height * RGB24)) {
	perror("Error writing image to pipe!");
	return 0;
    }
    
    return 1;
}


void fill_squares_data(struct square *squares, uint width, uint height) {

    int i;
    for (i = 0; i < n_squares; i++) {
	squares[i].big = (float) get_random(SQUARE_BIG_MAX, SQUARE_BIG_MIN);
	squares[i].x_pos = (float) get_random(width - (uint) squares[i].big, 5);
	squares[i].y_pos = (float) get_random(height - (uint) squares[i].big, 5);
	squares[i].colour = get_random(0xFFFFFF, 0x0);
	
	if ((get_random(2,0)) == 1)
	    squares[i].big_inc = FORWARD;
	else
	    squares[i].big_inc = BACKWARD;
	    
	if ((get_random(2,0)) == 1)
	    squares[i].x_direction = FORWARD;
	else
	    squares[i].x_direction = BACKWARD;

	if ((get_random(2,0)) == 1)
	    squares[i].y_direction = FORWARD;
	else
	    squares[i].y_direction = BACKWARD;
    }
}

int main (int argc, char **argv)
{
    if (argc < 3) {
	printf("\nInsufficient parameters.\n");
	printf("Usage: dummyfill [output] [widthxheight]\n\n");
	printf("example: videofake /dev/video0 352x288\n\n");
	
	exit(1);
    }
    
    uint width, height;
    sscanf(argv[2], "%dx%d", &width, &height);

    byte *image;
    image = (byte *) malloc(height * width * RGB24); // 24 Bits colour

    // Opening the device
    int devout;
    devout = open(argv[1], O_RDWR | O_NONBLOCK, 0);
    if (devout < 0) {
	perror ("Failed to open the specified device.");
	exit(1);
    }

    // Opening the output device
    if (start_pipe(devout, width, height)) {
	perror("Failed to initialize output pipe, exiting...");
	exit(1);
    }

    // Initializing random seed
    initialize_random_seed();
    
    // Allocating a random number of squares
    n_squares = get_random(SCREEN_MAX_SQUARES, SCREEN_MIN_SQUARES);
    struct square squares[n_squares];
    
    // Filling them
    fill_squares_data((struct square *) &squares, width, height);
    
    printf("Starting video stream.\n");
    while (1) {
	next_frame(image, width, height, (struct square *) &squares);
	if (put_image(devout, image, width, height) == 0)
	    exit(1);
		
	usleep(40); /* 25 FPS ?? */
    }

    // Exiting...
    close(devout);
    free(image);
    
    exit(0);
}
